# 텐서플로(TensorFlow)는 구글에서 개발한 오픈소스 딥러닝 라이브러리
# 머신러닝 모델을 설계, 학습, 평가 및 배포하는 데 사용
# 파이썬을 기반으로 사용하기 쉽고, 다양한 플랫폼과 호환되어 딥러닝 및 머신러닝 학습과 연구에 많이 활용

# 1단계: 환경 설정
# 필요한 라이브러리를 설치
# pip install tensorflow
# GPU 지원 텐서플로 설치
# pip install tensorflow-gpu

import tensorflow as tf
# 텐서플로 버전확인
# print("TensorFlow version:", tf.__version__)

# 텐서플로의 주요 구성 요소
# 데이터 파이프라인 구축 (tf.data): 데이터를 효과적으로 불러오고 처리할 수 있도록 돕는 모듈
# 케라스(Keras): 텐서플로에 포함된 고수준 API로, 신경망 모델을 쉽게 구축하고 훈련
# 텐서(Tensor): 텐서플로의 기본 데이터 구조로, 숫자의 다차원 배열
# 오퍼레이션(Operation): 텐서를 조작하고 연산하는 함수나 연산
# 그래프 및 세션(Graph and Session): 최신 버전의 텐서플로에서는 그래프와 세션 개념이 기본적으로 처리
# 그래프 구조로 모델을 표현하고 세션으로 실행하는 것이 기본 개념

# 텐서플로로 딥러닝 모델 구축하기
# 모델을 구축하는 기본 흐름은 데이터 준비, 데이터 전처리, 모델 설계, 모델 컴파일, 모델 학습, 모델 평가


# 데이터 시각화
# pip install matplotlib
import matplotlib.pyplot as plt

# 수학 계산과 배열 처리
import numpy as np


# 2단계: 데이터 준비
# 딥러닝 모델을 학습시키기 위해서는 데이터가 필요
# MNIST 데이터셋을 로드하고 데이터를 확인
# 텐서플로에서 제공하는 MNIST 데이터셋을 예제로 사용

# MNIST 데이터셋 로드
from tensorflow.keras.datasets import mnist

# mnist.load_data(): MNIST 데이터를 다운로드하고, 훈련 및 테스트 세트
# 데이터는 28x28 픽셀 크기의 흑백 이미지로 구성
(x_train, y_train), (x_test, y_test) = mnist.load_data()

# 데이터 크기 확인
# print(f"훈련 데이터 크기: {x_train.shape}, 레이블 크기: {y_train.shape}")
# print(f"테스트 데이터 크기: {x_test.shape}, 레이블 크기: {y_test.shape}")

# 샘플 데이터 시각화
# plt.imshow(x_train[0], cmap='gray')
# plt.title(f"Label: {y_train[0]}")
# plt.colorbar()
# plt.show()


# 3단계: 데이터 전처리
# CNN 모델은 (배치 크기, 높이, 너비, 채널) 형태의 데이터를 필요로 하므로 데이터 형태를 변환
# 데이터 정규화 (0~255 → 0~1)
# 정규화: 데이터의 값을 0~1 범위로 변환해 학습 속도를 높임
x_train = x_train.astype('float32') / 255.0
x_test = x_test.astype('float32') / 255.0

# CNN 입력 형태로 변환 (28x28 → 28x28x1)
x_train = x_train.reshape(-1, 28, 28, 1)
x_test = x_test.reshape(-1, 28, 28, 1)

# 레이블을 원-핫 인코딩 (0~9 -> 10차원 벡터)
# 원-핫 인코딩: 레이블(정답)을 딥러닝에서 처리 가능한 형태로 변환
from tensorflow.keras.utils import to_categorical
y_train = to_categorical(y_train, 10)
y_test = to_categorical(y_test, 10)

# print(f"훈련 데이터 크기: {x_train.shape}, 레이블 크기: {y_train.shape}")


# 4단계: 딥러닝 모델 구축
# CNN(합성곱 신경망)을 설계
# CNN은 이미지를 입력받아 특징을 추출하고, 마지막에 Dense Layer를 사용해 분류
# Keras API를 사용해 순차 모델(Sequential)을 구성
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2D, MaxPooling2D, Flatten, Dense, Dropout

# CNN 모델 설계
model = Sequential([
    Conv2D(32, (3, 3), activation='relu', input_shape=(28, 28, 1)),  # 합성곱 층
    MaxPooling2D(pool_size=(2, 2)),  # 풀링 층
    Dropout(0.25),  # 과적합 방지

    Conv2D(64, (3, 3), activation='relu'),  # 추가 합성곱 층
    MaxPooling2D(pool_size=(2, 2)),  # 추가 풀링 층
    Dropout(0.25),  # 추가 과적합 방지

    Flatten(),  # 1D로 펼침
    Dense(128, activation='relu'),  # 완전 연결층
    Dropout(0.5),  # 과적합 방지
    Dense(10, activation='softmax')  # 출력층 (10개의 클래스)
])

# 모델 요약 출력
# model.summary()


# 5단계: 모델 컴파일 및 학습
# 훈련 데이터를 사용해 모델을 학습
# 모델을 컴파일하고 훈련 데이터를 학습
# adam 옵티마이저와 categorical_crossentropy 손실 함수를 사용
# metrics: 모델 성능 측정 지표(accuracy: 정확도) 사용

# 모델 컴파일
model.compile(optimizer='adam',
              loss='categorical_crossentropy',
              metrics=['accuracy'])

# 모델 학습
history = model.fit(
    x_train, y_train,
    epochs=10,  # 학습 반복 횟수
    batch_size=128,  # 배치 크기
    validation_split=0.2  # 검증 데이터 비율
)
# epochs: 데이터셋을 몇 번 반복 학습할지 설정
# batch_size: 한 번에 학습할 데이터 샘플 수
# validation_split: 데이터의 일부를 검증 데이터로 분리


# 6단계: 학습 결과 시각화
# 학습 및 검증 손실과 정확도를 그래프로 시각화하여 모델의 학습 상태를 분석
# 학습 손실과 검증 손실 시각화
# plt.plot(history.history['loss'], label='Training Loss')
# plt.plot(history.history['val_loss'], label='Validation Loss')
# plt.title('Loss')
# plt.legend()
# plt.show()

# 학습 정확도와 검증 정확도 시각화
# plt.plot(history.history['accuracy'], label='Training Accuracy')
# plt.plot(history.history['val_accuracy'], label='Validation Accuracy')
# plt.title('Accuracy')
# plt.legend()
# plt.show()


# 7단계: 모델 평가 및 테스트
# 테스트 데이터를 사용하여 모델의 성능을 평가
# 샘플 데이터를 예측하고 결과를 확인
# 테스트 데이터 평가
test_loss, test_acc = model.evaluate(x_test, y_test)
print(f"테스트 손실: {test_loss:.4f}, 테스트 정확도: {test_acc:.4f}")

# 샘플 데이터 예측
predictions = model.predict(x_test[:5])

# 예측 결과 시각화
for i in range(5):
    plt.imshow(x_test[i].reshape(28, 28), cmap='gray')
    plt.title(f"True: {np.argmax(y_test[i])}, Predicted: {np.argmax(predictions[i])}")
    plt.axis('off')
    plt.show()


# MNIST 데이터셋에서 손글씨 숫자를 인식하는 딥러닝 모델을 구축하고 테스트하는 과정을 완료