# opencv-python 라이브러리 설치
import cv2
import numpy as np
# 데이터 시각화
# pip install matplotlib
import matplotlib.pyplot as plt

# load_model 함수를 사용하여 모델을 불러옴
from tensorflow.keras.models import load_model
loaded_model = load_model('./handwriting_num/cnn_handwriting_model.keras')


# DroidCam 스트림 URL 설정
ip_address = '192.168.0.8'  # DroidCam에서 확인한 IP 주소
port = '4747'  # DroidCam에서 사용하는 포트 (일반적으로 4747)
video_url = f'http://{ip_address}:{port}/video'

# OpenCV를 사용하여 스트림 열기
cap = cv2.VideoCapture(video_url)

if not cap.isOpened():
    print("스트림을 열 수 없습니다.")
    exit()


# 정사각형 이미지를 만드는 함수
def crop_to_square(image):
    """
    입력 이미지를 정사각형으로 자르기.
    짧은 변을 기준으로 중앙을 잘라 정사각형 생성.
    """
    h, w = image.shape[:2]
    # 짧은 쪽의 길이를 기준으로 중앙 영역 크기 결정
    size = min(h, w)
    start_x = (w - size) // 2  # 가로 중앙 계산
    start_y = (h - size) // 2  # 세로 중앙 계산
    return image[start_y:start_y + size, start_x:start_x + size]


# 키 입력 감지 및 명령 전송 루프
def control_car_sync():
    frame_counter = 0
    while True:
        # 프레임 읽기
        ret, frame = cap.read()
        if ret:

            # =====
            # 원본 이미지의 가로, 세로 크기
            (h, w) = frame.shape[:2]
            # 90도 회전하는 회전 행렬 생성
            center = (w // 2, h // 2)
            rotation_matrix = cv2.getRotationMatrix2D(center, -90, 1.0)
            # 회전 후 이미지 크기 계산
            cos = np.abs(rotation_matrix[0, 0])
            sin = np.abs(rotation_matrix[0, 1])
            # 새로운 너비와 높이 계산
            new_w = int((h * sin) + (w * cos))
            new_h = int((h * cos) + (w * sin))
            # 회전 행렬에 이동 보정 추가
            rotation_matrix[0, 2] += (new_w / 2) - center[0]
            rotation_matrix[1, 2] += (new_h / 2) - center[1]
            # 가로, 세로를 바꾼 크기로 warpAffine 적용 (확보된 크기로 변환)
            rotated_frame = cv2.warpAffine(frame, rotation_matrix, (new_w, new_h))
            # 회전된 이미지를 정사각형으로 잘라내기
            square_frame = crop_to_square(rotated_frame)
            # 회전된 정사각형 영상을 출력
            cv2.imshow('Rotated and Cropped Video', square_frame)
            # =====

            # 프레임을 화면에 표시
            # cv2.imshow('DroidCam Video', frame)
        else:
            print("프레임을 읽을 수 없습니다.")
            break

        # 이벤트 루프를 실행하면서 특정 시간 동안 키보드 입력을 기다리는 역할('q' 키를 누르면 종료)
        if cv2.waitKey(10) & 0xFF == ord('q'):
            break

        # 30 프레임마다 예측
        if frame_counter % 30 == 0:
            # RGB 이미지를 흑백으로 변환
            square_frame = cv2.cvtColor(square_frame, cv2.COLOR_BGR2GRAY)
            # 손글씨와 배경의 대비를 높이기 위해 이진화(Thresholding) 적용
            _, binary_image = cv2.threshold(square_frame, 128, 255, cv2.THRESH_BINARY)
            # 검은 글자가 흰 글자가 되도록 반전
            inverted = cv2.bitwise_not(binary_image)
            # 이미지를 28x28로 리사이즈 (이미 28x28인 경우 이 단계는 필요 없음)
            resized_frame = cv2.resize(inverted, (28, 28))

            # 회전된 정사각형 영상을 출력
            # 이미지를 2배로 확대 (INTER_NEAREST는 빠르게 처리)
            # cv2.resize를 사용한 크기 변경
            # fx=2와 fy=2로 각각 가로와 세로를 2배로 확대
            # interpolation=cv2.INTER_NEAREST는 픽셀 기반으로 단순히 값을 복제하여 빠르게 확대
            # 다양한 보간법 (Interpolation) 선택
            # cv2.INTER_NEAREST: 가장 빠르며 픽셀 기반 확대
            # cv2.INTER_LINEAR: 2배 또는 3배 확대에 적합
            # cv2.INTER_CUBIC: 선명한 이미지를 원할 때 사용(느림)
            scaled_frame = cv2.resize(resized_frame, (0, 0), fx=17, fy=17, interpolation=cv2.INTER_NEAREST)
            # 확대된 이미지를 출력
            cv2.imshow('Resize and Cropped Video', scaled_frame)

            # 데이터 정규화 (0~1 사이 값으로 변환)
            x_test = resized_frame.astype('float32') / 255.0
            # 데이터 형태 변환 (28x28 -> 28x28x1)
            x_test = x_test.reshape(1, 28, 28, 1)  # CNN 모델 입력 형태에 맞춤
            # 샘플 데이터 예측
            predictions = loaded_model.predict(x_test)
            # 예측 결과
            predicted_class = np.argmax(predictions)  # 가장 높은 확률을 가진 클래스
            confidence = predictions[0][predicted_class]  # 해당 클래스의 확률
            # 50% 이상일 경우에만 출력
            if confidence > 0.85:
                print(f"Predicted: {predicted_class}, Confidence: {confidence:.2f}")
            else:
                pass

        frame_counter += 1

    # 자원 해제
    cap.release()
    cv2.destroyAllWindows()


# 메인 실행 루프 (동기식으로 실행)
control_car_sync()